@@ -1,6 +1,24 @@
+#!/usr/bin/python
+#
+# Copyright (C) 2012 Yoav Aviram.
+#
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#
+#      http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
 import gdata.spreadsheet.service
 
 
+ID_FIELD = '__rowid__'
+
+
 class WorksheetException(Exception):
     """Base class for spreadsheet exceptions.
     """
@@ -92,9 +110,11 @@ class Worksheet(object):
         :param row:
             A row element.
         :return:
-            A dict.
+            A dictionary with rows.
         """
-        return dict([(key, row.custom[key].text) for key in row.custom])
+        result = dict([(key, row.custom[key].text) for key in row.custom])
+        result[ID_FIELD] = row.id.text.split('/')[-1]
+        return result
 
     def _get_row_entries(self, query=None):
         """Get Row Entries.
@@ -107,6 +127,23 @@ class Worksheet(object):
                 query=query, **self.keys).entry
         return self.entries
 
+    def _get_row_entry_by_id(self, id):
+        """Get Row Entry by ID
+
+        First search in cache, then fetch.
+        :param id:
+            A string row ID.
+        :return:
+            A row entry.
+        """
+        entry = [entry for entry in self._get_row_entries()
+                 if entry.id.text.split('/')[-1] == id]
+        if not entry:
+            entry = self.gd_client.GetListFeed(row_id=id, **self.keys).entry
+            if not entry:
+                raise WorksheetException("Row ID '{0}' not found.").format(id)
+        return entry[0]
+
     def _flush_cache(self):
         """Flush Entries Cache."""
         self.entries = None
@@ -131,7 +168,8 @@ class Worksheet(object):
         else:
             return None
 
-    def get_rows(self, query=None, order_by=None, reverse=None):
+    def get_rows(self, query=None, order_by=None,
+                 reverse=None, filter_func=None):
         """Get Rows
 
         :param query:
@@ -155,18 +193,51 @@ class Worksheet(object):
             A string which specifies whether to sort in descending or ascending
             order.Reverses default sort order: 'true' results in a descending
             sort; 'false' (the default) results in an ascending sort.
+        :param filter_func:
+            A lambda function which applied to each row, Gets a row dict as
+            argument and returns True or False. Used for filtering rows in
+            memory (as opposed to query which filters on the service side).
         :return:
-            A list of rows dictionaries.
+            A list of row dictionaries.
         """
         new_query = self._make_query(query, order_by, reverse)
-        if self.query != new_query:
+        if self.query is not None and self.query != new_query:
             self._flush_cache()
         self.query = new_query
-        return [self._row_to_dict(row)
+        rows = [self._row_to_dict(row)
             for row in self._get_row_entries(query=self.query)]
+        if filter_func:
+            rows = filter(filter_func, rows)
+        return rows
 
-    def update_row(self, index, row_data):
-        """Update Row
+    def update_row(self, row_data):
+        """Update Row (By ID).
+
+        Only the fields supplied will be updated.
+        :param row_data:
+            A dictionary containing row data. The row will be updated according
+            to the value in the ID_FIELD.
+        :return:
+            The updated row.
+        """
+        try:
+            id = row_data[ID_FIELD]
+        except KeyError:
+            raise WorksheetException("Row does not contain '{0}' field. "
+                                "Please update by index.".format(ID_FIELD))
+        entry = self._get_row_entry_by_id(id)
+        new_row = self._row_to_dict(entry)
+        new_row.update(row_data)
+        entry = self.gd_client.UpdateRow(entry, new_row)
+        if not isinstance(entry, gdata.spreadsheet.SpreadsheetsList):
+            raise WorksheetException("Row update failed: '{0}'".format(entry))
+        for i, e in enumerate(self.entries):
+            if e.id.text == entry.id.text:
+                self.entries[i] = entry
+        return self._row_to_dict(entry)
+
+    def update_row_by_index(self, index, row_data):
+        """Update Row By Index
 
         :param index:
             An integer designating the index of a row to update (zero based).
@@ -201,13 +272,30 @@ class Worksheet(object):
             self.entries.append(entry)
         return self._row_to_dict(entry)
 
-    def delete_row(self, index):
-        """Delete Row
+    def delete_row(self, row):
+        """Delete Row (By ID).
+
+        Requires that the given row dictionary contains an ID_FIELD.
+        :param row:
+            A row dictionary to delete.
+        """
+        try:
+            id = row[ID_FIELD]
+        except KeyError:
+            raise WorksheetException("Row does not contain '{0}' field. "
+                                "Please delete by index.".format(ID_FIELD))
+        entry = self._get_row_entry_by_id(id)
+        self.gd_client.DeleteRow(entry)
+        for i, e in enumerate(self.entries):
+            if e.id.text == entry.id.text:
+                del self.entries[i]
+
+    def delete_row_by_index(self, index):
+        """Delete Row By Index
 
         :param index:
-            A row index.
-            Index is relative to the returned result set, not to the original
-            spreadseet.
+            A row index. Index is relative to the returned result set, not to
+            the original spreadsheet.
         """
         entry = self._get_row_entries(self.query)[index]
         self.gd_client.DeleteRow(entry)
